/* This code is part of Freenet. It is distributed under the GNU General
 * Public License, version 2 (or at your option any later version). See
 * http://www.gnu.org/ for further details of the GPL. */
package freenet.client.async;

import java.io.IOException;

import com.db4o.ObjectContainer;

import freenet.client.ClientMetadata;
import freenet.client.InsertBlock;
import freenet.client.InsertContext;
import freenet.client.InsertException;
import freenet.client.Metadata;
import freenet.client.events.SplitfileProgressEvent;
import freenet.keys.BaseClientKey;
import freenet.keys.FreenetURI;
import freenet.node.RequestClient;
import freenet.node.RequestScheduler;
import freenet.support.Logger;
import freenet.support.SimpleFieldSet;
import freenet.support.api.Bucket;

public class ClientPutter extends BaseClientPutter implements PutCompletionCallback {

	final ClientCallback client;
	final Bucket data;
	final FreenetURI targetURI;
	final ClientMetadata cm;
	final InsertContext ctx;
	final String targetFilename;
	private ClientPutState currentState;
	private boolean finished;
	private final boolean getCHKOnly;
	private final boolean isMetadata;
	private boolean startedStarting;
	private final boolean binaryBlob;
	private FreenetURI uri;
	/** SimpleFieldSet containing progress information from last startup.
	 * Will be progressively cleared during startup. */
	private SimpleFieldSet oldProgress;

	/**
	 * @param client The object to call back when we complete, or don't.
	 * @param data The data to insert. This will not be freed by ClientPutter, the callback must do that. However,
	 * buckets used internally by the client layer will be freed.
	 * @param targetURI
	 * @param cm
	 * @param ctx
	 * @param scheduler
	 * @param priorityClass
	 * @param getCHKOnly
	 * @param isMetadata
	 * @param clientContext The client object for purposs of round-robin client balancing.
	 * @param stored The progress so far, stored as a SimpleFieldSet. Advisory; if there 
	 * is an error reading this in, we will restart from scratch.
	 * @param targetFilename If set, create a one-file manifest containing this filename pointing to this file.
	 */
	public ClientPutter(ClientCallback client, Bucket data, FreenetURI targetURI, ClientMetadata cm, InsertContext ctx,
			short priorityClass, boolean getCHKOnly, 
			boolean isMetadata, RequestClient clientContext, SimpleFieldSet stored, String targetFilename, boolean binaryBlob) {
		super(priorityClass, clientContext);
		this.cm = cm;
		this.isMetadata = isMetadata;
		this.getCHKOnly = getCHKOnly;
		this.client = client;
		this.data = data;
		this.targetURI = targetURI;
		this.ctx = ctx;
		this.finished = false;
		this.cancelled = false;
		this.oldProgress = stored;
		this.targetFilename = targetFilename;
		this.binaryBlob = binaryBlob;
	}

	public void start(boolean earlyEncode, ObjectContainer container, ClientContext context) throws InsertException {
		start(earlyEncode, false, container, context);
	}
	
	public boolean start(boolean earlyEncode, boolean restart, ObjectContainer container, ClientContext context) throws InsertException {
		if(persistent())
			container.activate(client, 1);
		if(Logger.shouldLog(Logger.MINOR, this))
			Logger.minor(this, "Starting "+this);
		try {
			this.targetURI.checkInsertURI();
			
			if(data == null)
				throw new InsertException(InsertException.BUCKET_ERROR, "No data to insert", null);
			
			boolean cancel = false;
			synchronized(this) {
				if(restart) {
					if(currentState != null && !finished) return false;
					finished = false;
				}
				if(startedStarting) return false;
				startedStarting = true;
				if(currentState != null) return false;
				cancel = this.cancelled;
				if(!cancel) {
					if(!binaryBlob) {
						ClientMetadata meta = cm;
						if(meta != null) meta = persistent() ? meta.clone() : meta; 
						currentState =
							new SingleFileInserter(this, this, new InsertBlock(data, meta, persistent() ? targetURI.clone() : targetURI), isMetadata, ctx, 
									false, getCHKOnly, false, null, null, false, targetFilename, earlyEncode);
					} else
						currentState =
							new BinaryBlobInserter(data, this, null, false, priorityClass, ctx, context, container);
				}
			}
			if(cancel) {
				onFailure(new InsertException(InsertException.CANCELLED), null, container, context);
				oldProgress = null;
				return false;
			}
			synchronized(this) {
				cancel = cancelled;
			}
			if(cancel) {
				onFailure(new InsertException(InsertException.CANCELLED), null, container, context);
				oldProgress = null;
				if(persistent())
					container.store(this);
				return false;
			}
			if(Logger.shouldLog(Logger.MINOR, this))
				Logger.minor(this, "Starting insert: "+currentState);
			if(currentState instanceof SingleFileInserter)
				((SingleFileInserter)currentState).start(oldProgress, container, context);
			else
				currentState.schedule(container, context);
			synchronized(this) {
				oldProgress = null;
				cancel = cancelled;
			}
			if(persistent()) {
				container.store(this);
				// It has scheduled, we can safely deactivate it now, so it won't hang around in memory.
				container.deactivate(currentState, 1);
			}
			if(cancel) {
				onFailure(new InsertException(InsertException.CANCELLED), null, container, context);
				return false;
			}
		} catch (InsertException e) {
			Logger.error(this, "Failed to start insert: "+e, e);
			synchronized(this) {
				finished = true;
				oldProgress = null;
				currentState = null;
			}
			if(persistent())
				container.store(this);
			// notify the client that the insert could not even be started
			if (this.client!=null) {
				this.client.onFailure(e, this, container);
			}
		} catch (IOException e) {
			Logger.error(this, "Failed to start insert: "+e, e);
			synchronized(this) {
				finished = true;
				oldProgress = null;
				currentState = null;
			}
			if(persistent())
				container.store(this);
			// notify the client that the insert could not even be started
			if (this.client!=null) {
				this.client.onFailure(new InsertException(InsertException.BUCKET_ERROR, e, null), this, container);
			}
		} catch (BinaryBlobFormatException e) {
			Logger.error(this, "Failed to start insert: "+e, e);
			synchronized(this) {
				finished = true;
				oldProgress = null;
				currentState = null;
			}
			if(persistent())
				container.store(this);
			// notify the client that the insert could not even be started
			if (this.client!=null) {
				this.client.onFailure(new InsertException(InsertException.BINARY_BLOB_FORMAT_ERROR, e, null), this, container);
			}
		} 
		if(Logger.shouldLog(Logger.MINOR, this))
			Logger.minor(this, "Started "+this);
		return true;
	}

	public void onSuccess(ClientPutState state, ObjectContainer container, ClientContext context) {
		if(persistent())
			container.activate(client, 1);
		ClientPutState oldState;
		synchronized(this) {
			finished = true;
			oldState = currentState;
			currentState = null;
			oldProgress = null;
		}
		if(oldState != null && persistent()) {
			container.activate(oldState, 1);
			oldState.removeFrom(container, context);
		}
		if(state != null && state != oldState && persistent())
			state.removeFrom(container, context);
		if(super.failedBlocks > 0 || super.fatallyFailedBlocks > 0 || super.successfulBlocks < super.totalBlocks) {
			Logger.error(this, "Failed blocks: "+failedBlocks+", Fatally failed blocks: "+fatallyFailedBlocks+
					", Successful blocks: "+successfulBlocks+", Total blocks: "+totalBlocks+" but success?! on "+this+" from "+state,
					new Exception("debug"));
		}
		if(persistent())
			container.store(this);
		client.onSuccess(this, container);
	}

	public void onFailure(InsertException e, ClientPutState state, ObjectContainer container, ClientContext context) {
		if(Logger.shouldLog(Logger.MINOR, this)) Logger.minor(this, "onFailure() for "+this+" : "+state+" : "+e, e);
		if(persistent())
			container.activate(client, 1);
		ClientPutState oldState;
		synchronized(this) {
			finished = true;
			oldState = currentState;
			currentState = null;
			oldProgress = null;
		}
		if(oldState != null && persistent()) {
			container.activate(oldState, 1);
			oldState.removeFrom(container, context);
		}
		if(state != null && state != oldState && persistent())
			state.removeFrom(container, context);
		if(persistent())
			container.store(this);
		client.onFailure(e, this, container);
	}

	@Override
	public void onMajorProgress(ObjectContainer container) {
		if(persistent())
			container.activate(client, 1);
		client.onMajorProgress(container);
	}
	
	public void onEncode(BaseClientKey key, ClientPutState state, ObjectContainer container, ClientContext context) {
		if(persistent())
			container.activate(client, 1);
		synchronized(this) {
			if(this.uri != null) {
				Logger.error(this, "onEncode() called twice? Already have a uri: "+uri+" for "+this);
				if(persistent())
					this.uri.removeFrom(container);
			}
			this.uri = key.getURI();
			if(targetFilename != null)
				uri = uri.pushMetaString(targetFilename);
		}
		if(persistent())
			container.store(this);
		client.onGeneratedURI(uri, this, container);
	}

	@Override
	public void cancel(ObjectContainer container, ClientContext context) {
		if(Logger.shouldLog(Logger.MINOR, this))
			Logger.minor(this, "Cancelling "+this, new Exception("debug"));
		ClientPutState oldState = null;
		synchronized(this) {
			if(cancelled) return;
			if(finished) return;
			super.cancel();
			oldState = currentState;
		}
		if(persistent()) {
			container.store(this);
			if(oldState != null)
				container.activate(oldState, 1);
		}
		if(oldState != null) oldState.cancel(container, context);
		onFailure(new InsertException(InsertException.CANCELLED), null, container, context);
	}
	
	@Override
	public synchronized boolean isFinished() {
		return finished || cancelled;
	}

	@Override
	public FreenetURI getURI() {
		return uri;
	}

	public void onTransition(ClientPutState oldState, ClientPutState newState, ObjectContainer container) {
		if(newState == null) throw new NullPointerException();
		
		// onTransition is *not* responsible for removing the old state, the caller is.
		synchronized (this) {
			if (currentState == oldState) {
				currentState = newState;
				if(persistent())
					container.store(this);
				return;
			}
		}
		Logger.error(this, "onTransition: cur=" + currentState + ", old=" + oldState + ", new=" + newState);
	}

	public void onMetadata(Metadata m, ClientPutState state, ObjectContainer container, ClientContext context) {
		Logger.error(this, "Got metadata on "+this+" from "+state+" (this means the metadata won't be inserted)");
	}
	
	@Override
	public void notifyClients(ObjectContainer container, ClientContext context) {
		if(persistent())
			container.activate(ctx, 2);
		ctx.eventProducer.produceEvent(new SplitfileProgressEvent(this.totalBlocks, this.successfulBlocks, this.failedBlocks, this.fatallyFailedBlocks, this.minSuccessBlocks, this.blockSetFinalized), container, context);
	}
	
	public void onBlockSetFinished(ClientPutState state, ObjectContainer container, ClientContext context) {
		if(Logger.shouldLog(Logger.MINOR, this))
			Logger.minor(this, "Set finished", new Exception("debug"));
		blockSetFinalized(container, context);
	}

	public void onFetchable(ClientPutState state, ObjectContainer container) {
		if(persistent())
			container.activate(client, 1);
		client.onFetchable(this, container);
	}

	public boolean canRestart() {
		if(currentState != null && !finished) {
			Logger.minor(this, "Cannot restart because not finished for "+uri);
			return false;
		}
		if(data == null) return false;
		return true;
	}

	public boolean restart(boolean earlyEncode, ObjectContainer container, ClientContext context) throws InsertException {
		return start(earlyEncode, true, container, context);
	}

	@Override
	public void onTransition(ClientGetState oldState, ClientGetState newState, ObjectContainer container) {
		// Ignore, at the moment
	}

	@Override
	public void removeFrom(ObjectContainer container, ClientContext context) {
		container.activate(cm, 2);
		cm.removeFrom(container);
		// This is passed in. We should not remove it, because the caller (ClientPutBase) should remove it.
//		container.activate(ctx, 1);
//		ctx.removeFrom(container);
		container.activate(targetURI, 5);
		targetURI.removeFrom(container);
		if(uri != null) {
			container.activate(uri, 5);
			uri.removeFrom(container);
		}
		super.removeFrom(container, context);
	}
}
